/*
 * Show lines from a file.
 *
 * Copyright (C) 2022, 2023, 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>



/* Read a line from the file. */

static int readline(FILE *file, int *lineno, char **start, char **end)
{
  static char buf[256];
  static offset_t current = 0, limit = 0;
  static int next_lineno = 1;
  char *newline;

  /* Advance to the next line if previously found. */

  if (next_lineno > *lineno)
    *lineno = next_lineno;

  /* Obtain file content. */

  if (!limit)
  {
    limit = fread(buf, sizeof(char), 256, file);

    if (!limit)
    {
      *start = NULL;
      *end = NULL;
      return 0;
    }
  }

  /* Find newline. */

  newline = (char *) memchr(buf + current, (int) '\n', limit - current);
  *start = (buf + current);

  /* Return final line fragment, potentially leaving more available text. */

  if (newline != NULL)
  {
    /* Advance the line number when reading again. */

    next_lineno = (*lineno) + 1;

    /* Define the end of the line and the start of the next line. */

    *end = newline + 1;
    current = *end - buf;

    if (current >= limit)
    {
      current = 0;
      limit = 0;
    }

    return 1;
  }

  /* Or return mid-line fragment reaching the limit of the available text. */

  else
  {
    *end = buf + limit;
    current = 0;
    limit = 0;

    return 1;
  }
}



int main(int argc, char *argv[])
{
  FILE *file;
  int lineno, startline, numlines;
  char *start, *end;

  if (argc < 4)
    return 1;

  if (!strcmp(argv[1], "-"))
    file = stdin;
  else
    file = fopen(argv[1], "r");

  if (file == NULL)
  {
    fprintf(stdout, "Could not open file.\n");
    return 1;
  }

  startline = atoi(argv[2]);
  numlines = atoi(argv[3]);

  lineno = 1;

  while (lineno < startline + numlines)
  {
    /* Read a line, updating the current line number. */

    if (!readline(file, &lineno, &start, &end))
    {
      /* If the last line was passed, an empty line is an acceptable end-of-file
         condition. */

      if (lineno >= startline + numlines)
        break;

      /* Otherwise, indicate that end-of-file occurred. */

      fprintf(stdout, "EOF error at line %d.\n", lineno);
      return 1;
    }

    /* Emit line content while within the desired range of lines. */

    if ((lineno >= startline) && (lineno < startline + numlines))
      fwrite(start, sizeof(char), end - start, stdout);
  }

  fclose(file);

  return 0;
}

/* vim: tabstop=2 expandtab shiftwidth=2
*/
